# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
Controlling visual aspects of Yuio with themes.

Theme base class
----------------

The overall look and feel of a Yuio application is declared
in a :class:`Theme` object:

.. autoclass:: Theme

    .. autoattribute:: progress_bar_width

    .. autoattribute:: progress_bar_start_symbol

    .. autoattribute:: progress_bar_end_symbol

    .. autoattribute:: progress_bar_done_symbol

    .. autoattribute:: progress_bar_pending_symbol

    .. autoattribute:: spinner_pattern

    .. autoattribute:: spinner_update_rate_ms

    .. autoattribute:: spinner_static_symbol

    .. autoattribute:: msg_decorations

    .. automethod:: set_msg_decoration

    .. automethod:: _set_msg_decoration_if_not_overridden

    .. autoattribute:: colors

    .. automethod:: set_color

    .. automethod:: _set_color_if_not_overridden

    .. automethod:: get_color

    .. automethod:: to_color


Default theme
-------------

Use the following loader to create an instance of the default theme:

.. autofunction:: load

.. autoclass:: DefaultTheme

"""

from __future__ import annotations

import dataclasses
import functools
import os
from dataclasses import dataclass

import yuio.term
from yuio import _typing as _t

__all__ = [
    "Theme",
    "DefaultTheme",
    "load",
]

T = _t.TypeVar("T")


class _ImmutableDictProxy(_t.Mapping[str, T], _t.Generic[T]):
    def __init__(self, data: dict[str, T], /, *, attr: str):
        self.__data = data
        self.__attr = attr

    def items(self) -> _t.ItemsView[str, T]:
        return self.__data.items()

    def keys(self) -> _t.KeysView[str]:
        return self.__data.keys()

    def values(self) -> _t.ValuesView[T]:
        return self.__data.values()

    def __len__(self):
        return len(self.__data)

    def __getitem__(self, key):
        return self.__data[key]

    def __iter__(self):
        return iter(self.__data)

    def __contains__(self, key):
        return key in self.__data

    def __repr__(self):
        return repr(self.__data)

    def __setitem__(self, key, item):
        raise RuntimeError(f"Theme.{self.__attr} is immutable")

    def __delitem__(self, key):
        raise RuntimeError(f"Theme.{self.__attr} is immutable")


class Theme:
    """
    Base class for Yuio themes.

    """

    msg_decorations: _t.Mapping[str, str] = {
        "heading/section": "",
        "heading/1": "⣿ ",
        "heading/2": "",
        "heading/3": "",
        "heading/4": "",
        "heading/5": "",
        "heading/6": "",
        "question": "> ",
        "task": "> ",
        "thematic_break": "╌╌╌╌╌",
        "list": "•   ",
        "quote": ">   ",
        "code": " " * 8,
        # TODO: support these in widgets
        # 'menu_selected_item': '▶︎',
        # 'menu_default_item': '★',
        # 'menu_select': '#',
        # 'menu_search': '/',
    }
    """
    Decorative symbols for certain text elements, such as headings,
    list items, etc.

    This mapping becomes immutable once a theme class is created. The only possible
    way to modify it is by using :meth:`~Theme.set_msg_decoration`
    or :meth:`~Theme._set_msg_decoration_if_not_overridden`.

    """

    __msg_decorations: dict[str, str]
    """
    An actual mutable version of :attr:`~Theme.msg_decorations`
    is kept here, because `__init_subclass__` will replace
    :attr:`~Theme.msg_decorations` with an immutable proxy.

    """

    __msg_decoration_sources: dict[str, type | None] = {}
    """
    Keeps track of where a message decoration was inherited from. This var is used
    to avoid `__init__`-ing message decorations that were overridden in a subclass.

    """

    progress_bar_width: int = 15
    """
    Width of a progress bar for :class:`yuio.io.Task`.

    """

    progress_bar_start_symbol: str = ""
    """
    A symbol rendered on a left side of a progressbar.

    Set to ``"["`` to enclose progressbar in square brackets, for example.

    """

    progress_bar_end_symbol: str = ""
    """
    A symbol rendered on a right side of a progressbar.

    Set to ``"]"`` to enclose progressbar in square brackets, for example.

    """

    progress_bar_done_symbol: str = "■"
    """
    Symbol rendered in the filled portion of a progressbar.

    """

    progress_bar_pending_symbol: str = "□"
    """
    Symbol rendered in the unfilled portion of a progressbar.

    """

    spinner_pattern: _t.Sequence[str] = "⣤⣤⣤⠶⠛⠛⠛⠶"
    """
    Spinner pattern for running tasks that don't have a progressbar.

    Every tick, a symbol in front of a task's heading updates, showing elements
    of this sequence.

    """

    spinner_update_rate_ms: int = 200
    """
    How often the :attr:`~Theme.spinner_pattern` changes.

    """

    spinner_static_symbol = "⣿"
    """
    Symbol for finished and failed tasks.

    It meant to resemble a static spinner.

    """

    colors: _t.Mapping[str, str | yuio.term.Color | list[str | yuio.term.Color]] = {
        "code": "magenta",
        "note": "green",
        "bold": yuio.term.Color.STYLE_BOLD,
        "b": "bold",
        "dim": yuio.term.Color.STYLE_DIM,
        "d": "dim",
        "normal": yuio.term.Color.FORE_NORMAL,
        "normal_dim": yuio.term.Color.FORE_NORMAL_DIM,
        "red": yuio.term.Color.FORE_RED,
        "green": yuio.term.Color.FORE_GREEN,
        "yellow": yuio.term.Color.FORE_YELLOW,
        "blue": yuio.term.Color.FORE_BLUE,
        "magenta": yuio.term.Color.FORE_MAGENTA,
        "cyan": yuio.term.Color.FORE_CYAN,
    }
    """
    Mapping of color paths to actual colors.

    Themes use color paths to describe styles and colors for different
    parts of an application. yuio.term.Color paths are similar to file paths,
    they use snake case identifiers separated by slashes, and consist of
    two parts separated by a colon.

    The first part represents an object, i.e. what are we coloring.

    The second part represents a context, i.e. what is the state or location
    of an object that we're coloring.

    For example, a color for the filled part of the task's progress bar
    has path ``"task/progressbar/done"``, a color for a text of an error
    log record has path ``"log/message:error"``, and a color for a string escape
    sequence in a highlighted python code has path ``"hl/str/esc:python"``.

    A color at a certain path is propagated to all sub-paths. For example,
    if ``"task/progressbar"`` is bold, and ``"task/progressbar/done"`` is green,
    the final color will be bold green.

    Each color path can be associated with either an instance
    of :class:`~yuio.term.Color`, another path, or a list of colors and paths.

    If path is mapped to a :class:`~yuio.term.Color`, then the path is associated
    with that particular color.

    If path is mapped to another path, then the path is associated with
    the color value for that other path (please don't create recursions here).

    If path is mapped to a list of colors and paths, then those colors and paths
    are combined.

    For example::

        colors = {
            'heading_color': yuio.term.Color.BOLD,
            'error_color': yuio.term.Color.RED,
            'tb/heading': ['heading_color', 'error_color'],
        }

    Here, color of traceback's heading ``"tb/heading"`` will be bold and red.

    The base theme class provides colors for basic tags, such as `bold`, `red`,
    `code`, `note`, etc. :class:`DefaultTheme` expands on it, providing main
    colors that control the overall look of the theme, and then colors for all
    interface elements.

    When deriving from a theme, you can override this mapping. When looking up
    colors via :meth:`~Theme.get_color`, base classes will be tried for color,
    in order of method resolution.

    This mapping becomes immutable once a theme class is created. The only possible
    way to modify it is by using :meth:`~Theme.set_color`
    or :meth:`~Theme._set_color_if_not_overridden`.

    """

    __colors: dict[str, str | yuio.term.Color | list[str | yuio.term.Color]]
    """
    An actual mutable version of :attr:`~Theme.colors`
    is kept here, because `__init_subclass__` will replace
    :attr:`~Theme.colors` with an immutable proxy.

    """

    __color_sources: dict[str, type | None] = {}
    """
    Keeps track of where a color was inherited from. This var is used
    to avoid `__init__`-ing colors that were overridden in a subclass.

    """

    __expected_source: type | None = None
    """
    When running an `__init__` function, this variable will be set to the class
    that implemented it, regardless of type of `self`.

    That is, inside `DefaultTheme.__init__`, `__expected_source` is set
    to `DefaultTheme`, in `MyTheme.__init__` it is `MyTheme`, etc.

    This is possible because `__init_subclass__` wraps any implementation
    of `__init__` into a wrapper that sets this variable.

    """

    def __init__(self):
        pass

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        colors = {}
        color_sources = {}
        for base in reversed(cls.__mro__):
            base_colors = getattr(base, "_Theme__colors", {})
            colors.update(base_colors)
            base_color_sources = getattr(base, "_Theme__color_sources", {})
            color_sources.update(base_color_sources)

        colors.update(cls.colors)
        color_sources.update(dict.fromkeys(cls.colors.keys(), cls))

        cls.__colors = colors
        cls.__color_sources = color_sources
        cls.colors = _ImmutableDictProxy(cls.__colors, attr="colors")

        msg_decorations = {}
        msg_decoration_sources = {}
        for base in reversed(cls.__mro__):
            base_msg_decorations = getattr(base, "_Theme__msg_decorations", {})
            msg_decorations.update(base_msg_decorations)
            base_msg_decoration_sources = getattr(
                base, "_Theme__msg_decoration_sources", {}
            )
            msg_decoration_sources.update(base_msg_decoration_sources)

        msg_decorations.update(cls.msg_decorations)
        msg_decoration_sources.update(dict.fromkeys(cls.msg_decorations.keys(), cls))

        cls.__msg_decorations = msg_decorations
        cls.__msg_decoration_sources = msg_decoration_sources
        cls.msg_decorations = _ImmutableDictProxy(
            cls.__msg_decorations, attr="msg_decorations"
        )

        if init := cls.__dict__.get("__init__", None):

            @functools.wraps(init)
            def _wrapped_init(_self, *args, **kwargs):
                prev_expected_source = _self._Theme__expected_source
                _self._Theme__expected_source = cls
                try:
                    return init(_self, *args, **kwargs)
                finally:
                    _self._Theme__expected_source = prev_expected_source

            cls.__init__ = _wrapped_init  # type: ignore

    def _set_msg_decoration_if_not_overridden(
        self,
        name: str,
        msg_decoration: str,
        /,
    ):
        """
        Set message decoration by name, but only if it wasn't overridden
        in a subclass.

        This method should be called from `__init__` implementations
        to dynamically set message decorations. It will only set the decoration
        if it was not overridden by any child class.

        """

        if self.__expected_source is None:
            raise RuntimeError(
                f"_set_msg_decoration_if_not_overridden should only be called from __init__"
            )
        source = self.__msg_decoration_sources.get(name, Theme)
        # The class that's `__init__` is currently running should be a parent
        # of the msg_decoration's source. This means that the msg_decoration was assigned by a parent.
        if source is not None and issubclass(self.__expected_source, source):
            self.set_msg_decoration(name, msg_decoration)

    def set_msg_decoration(
        self,
        name: str,
        msg_decoration: str,
        /,
    ):
        """
        Set message decoration by name.

        """

        if "_Theme__msg_decorations" not in self.__dict__:
            self.__msg_decorations = self.__class__.__msg_decorations.copy()
            self.__msg_decoration_sources = (
                self.__class__.__msg_decoration_sources.copy()
            )
            self.msg_decorations = _ImmutableDictProxy(
                self.__msg_decorations, attr="msg_decorations"
            )
        self.__msg_decorations[name] = msg_decoration
        self.__msg_decoration_sources[name] = self.__expected_source

    def _set_color_if_not_overridden(
        self,
        path: str,
        color: str | yuio.term.Color | list[str | yuio.term.Color],
        /,
    ):
        """
        Set color by path, but only if the color was not overridden in a subclass.

        This method should be called from `__init__` implementations
        to dynamically set colors. It will only set the color if it was not overridden
        by any child class.

        """

        if self.__expected_source is None:
            raise RuntimeError(
                f"_set_color_if_not_overridden should only be called from __init__"
            )
        source = self.__color_sources.get(path, Theme)
        # The class who's `__init__` is currently running should be a parent
        # of the color's source. This means that the color was assigned by a parent.
        if source is not None and issubclass(self.__expected_source, source):
            self.set_color(path, color)

    def set_color(
        self,
        path: str,
        color: str | yuio.term.Color | list[str | yuio.term.Color],
        /,
    ):
        """
        Set color by path.

        """

        if "_Theme__colors" not in self.__dict__:
            self.__colors = self.__class__.__colors.copy()
            self.__color_sources = self.__class__.__color_sources.copy()
            self.colors = _ImmutableDictProxy(self.__colors, attr="colors")
        self.__colors[path] = color
        self.__color_sources[path] = self.__expected_source
        self.get_color.cache_clear()
        self.__dict__.pop("_Theme__color_tree", None)

    @dataclass(**yuio._with_slots())
    class __ColorTree:
        """
        Prefix-like tree that contains all of the theme's colors.

        """

        colors: str | yuio.term.Color | list[str | yuio.term.Color] = (
            yuio.term.Color.NONE
        )
        """
        Colors in this node.

        """

        loc: dict[str, Theme.__ColorTree] = dataclasses.field(default_factory=dict)
        """
        Location part of the tree.

        """

        ctx: dict[str, Theme.__ColorTree] = dataclasses.field(default_factory=dict)
        """
        Context part of the tree.

        """

    @functools.cached_property
    def __color_tree(self) -> Theme.__ColorTree:
        root = self.__ColorTree()

        for path, colors in self.__colors.items():
            loc, ctx = self.__parse_path(path)

            node = root

            for part in loc:
                if part not in node.loc:
                    node.loc[part] = self.__ColorTree()
                node = node.loc[part]

            for part in ctx:
                if part not in node.ctx:
                    node.ctx[part] = self.__ColorTree()
                node = node.ctx[part]

            node.colors = colors

        return root

    @staticmethod
    def __parse_path(path: str, /) -> tuple[list[str], list[str]]:
        path_parts = path.split(":", maxsplit=1)
        if len(path_parts) == 1:
            loc, ctx = path_parts[0], ""
        else:
            loc, ctx = path_parts
        return loc.split("/") if loc else [], ctx.split("/") if ctx else []

    @_t.final
    @functools.cache
    def get_color(self, path: str, /) -> yuio.term.Color:
        """
        Lookup a color by path.

        """

        loc, ctx = self.__parse_path(path)
        return self.__get_color_in_loc(self.__color_tree, loc, ctx)

    def __get_color_in_loc(
        self, node: Theme.__ColorTree, loc: list[str], ctx: list[str]
    ):
        color = yuio.term.Color.NONE

        for part in loc:
            if part not in node.loc:
                break
            color |= self.__get_color_in_ctx(node, ctx)
            node = node.loc[part]

        return color | self.__get_color_in_ctx(node, ctx)

    def __get_color_in_ctx(self, node: Theme.__ColorTree, ctx: list[str]):
        color = yuio.term.Color.NONE

        for part in ctx:
            if part not in node.ctx:
                break
            color |= self.__get_color_in_node(node)
            node = node.ctx[part]

        return color | self.__get_color_in_node(node)

    def __get_color_in_node(self, node: Theme.__ColorTree) -> yuio.term.Color:
        color = yuio.term.Color.NONE

        if isinstance(node.colors, str):
            color |= self.get_color(node.colors)
        elif isinstance(node.colors, list):
            for c in node.colors:
                color |= self.get_color(c) if isinstance(c, str) else c
        else:
            color |= node.colors

        return color

    def to_color(self, color_or_path: yuio.term.Color | str | None) -> yuio.term.Color:
        """
        Convert color or color path to color.

        """

        if color_or_path is None:
            return yuio.term.Color.NONE
        elif isinstance(color_or_path, yuio.term.Color):
            return color_or_path
        else:
            return self.get_color(color_or_path)


Theme.__init_subclass__()


class DefaultTheme(Theme):
    """
    Default Yuio theme. Adapts for terminal background color,
    if one can be detected.

    This theme defines *main colors*, which you can override by subclassing.

    - ``"heading_color"``: for headings,
    - ``"primary_color"``: for main text,
    - ``"accent_color"``: for visually highlighted elements,
    - ``"secondary_color"``: for visually dimmed elements,
    - ``"error_color"``: for everything that indicates an error,
    - ``"warning_color"``: for everything that indicates a warning,
    - ``"success_color"``: for everything that indicates a success,
    - ``"low_priority_color_a"``: for auxiliary elements such as help widget,
    - ``"low_priority_color_b"``: for auxiliary elements such as help widget,
      even lower priority.

    """

    colors = {
        #
        # Main settings
        # -------------
        # This section controls the overall theme look.
        # Most likely you'll want to change accent colors from here.
        "heading_color": ["bold", "primary_color"],
        "primary_color": "normal",
        "accent_color": "magenta",
        "accent_color_2": "cyan",
        "secondary_color": "normal_dim",
        "error_color": "red",
        "warning_color": "yellow",
        "success_color": "green",
        "low_priority_color_a": "normal_dim",
        "low_priority_color_b": "normal_dim",
        #
        # Common tags
        # -----------
        "code": "accent_color",
        "note": "accent_color_2",
        "path": "accent_color",
        #
        # IO messages and text
        # --------------------
        "msg/decoration": "secondary_color",
        "msg/decoration:heading": "accent_color",
        "msg/decoration:thematic_break": "secondary_color",
        "msg/text": "primary_color",
        "msg/text:heading": "heading_color",
        "msg/text:heading/section": "accent_color",
        "msg/text:question": "heading_color",
        "msg/text:error": "error_color",
        "msg/text:warning": "warning_color",
        "msg/text:success": ["bold", "success_color"],
        "msg/text:failure": ["bold", "error_color"],
        "msg/text:info": "primary_color",
        "msg/text:thematic_break": "secondary_color",
        #
        # Log messages
        # ------------
        "log/asctime": "secondary_color",
        "log/logger": "secondary_color",
        "log/level": "heading_color",
        "log/level:critical": "error_color",
        "log/level:error": "error_color",
        "log/level:warning": "warning_color",
        "log/level:info": "success_color",
        "log/level:debug": "secondary_color",
        "log/message": "primary_color",
        #
        # Tasks and progress bars
        # -----------------------
        "task": "secondary_color",
        "task/decoration:running": "accent_color",
        "task/decoration:done": "success_color",
        "task/decoration:error": "error_color",
        "task/progressbar/done": "accent_color",
        "task/progressbar/pending": "secondary_color",
        "task/heading": "heading_color",
        "task/progress": "secondary_color",
        "task/comment": "primary_color",
        #
        # Syntax highlighting
        # -------------------
        "hl/kwd": "bold",
        "hl/str": yuio.term.Color.NONE,
        "hl/str/esc": "bold",
        "hl/lit": yuio.term.Color.NONE,
        "hl/punct": "blue",
        "hl/punct:sh-usage": "secondary_color",
        "hl/comment": "secondary_color",
        "hl/prog": "bold",
        "hl/flag": "accent_color_2",
        "hl/meta:diff": "accent_color",
        "hl/added:diff": "green",
        "hl/removed:diff": "red",
        "hl/metavar": "bold",
        "tb/heading": ["bold", "red"],
        "tb/message": "tb/heading",
        "tb/frame/usr/file/module": "code",
        "tb/frame/usr/file/line": "code",
        "tb/frame/usr/file/path": "code",
        "tb/frame/usr/code": yuio.term.Color.NONE,
        "tb/frame/usr/highlight": "low_priority_color_a",
        "tb/frame/lib": "dim",
        "tb/frame/lib/file/module": "tb/frame/usr/file/module",
        "tb/frame/lib/file/line": "tb/frame/usr/file/line",
        "tb/frame/lib/file/path": "tb/frame/usr/file/path",
        "tb/frame/lib/code": "tb/frame/usr/code",
        "tb/frame/lib/highlight": "tb/frame/usr/highlight",
        #
        # Menu and widgets
        # ----------------
        "menu/text": "primary_color",
        "menu/text/heading": ["menu/text", "heading_color"],
        "menu/text/help_info:help": "low_priority_color_a",
        "menu/text/help_msg:help": "low_priority_color_b",
        "menu/text/help_key:help": "low_priority_color_a",
        "menu/text/help_sep:help": "low_priority_color_b",
        "menu/text/help_key:help_menu": "accent_color_2",
        "menu/text/help_sep:help_menu": "secondary_color",
        "menu/text/esc": "accent_color",
        "menu/text/comment": "accent_color_2",
        "menu/text/comment/decoration": "secondary_color",
        "menu/text:choice/active": "accent_color",
        "menu/text:choice/active/selected": ["bold"],
        "menu/text:choice/normal/selected": ["accent_color_2", "bold"],
        "menu/text:choice/normal/dir": "blue",
        "menu/text:choice/normal/exec": "red",
        "menu/text:choice/normal/symlink": "magenta",
        "menu/text:choice/normal/socket": "green",
        "menu/text:choice/normal/pipe": "yellow",
        "menu/text:choice/normal/block_device": ["cyan", "bold"],
        "menu/text:choice/normal/char_device": ["yellow", "bold"],
        "menu/text/comment:choice/normal/original": "success_color",
        "menu/text/comment:choice/normal/corrected": "error_color",
        "menu/text/prefix:choice/normal": "primary_color",
        "menu/text/prefix:choice/normal/selected": ["accent_color_2", "bold"],
        "menu/text/prefix:choice/active": "accent_color",
        "menu/text/prefix:choice/active/selected": ["bold"],
        "menu/text/suffix:choice/normal": "primary_color",
        "menu/text/suffix:choice/normal/selected": ["accent_color_2", "bold"],
        "menu/text/suffix:choice/active": "accent_color",
        "menu/text/suffix:choice/active/selected": ["bold"],
        "menu/text:choice/status_line": "low_priority_color_b",
        "menu/text:choice/status_line/number": "low_priority_color_a",
        "menu/placeholder": "secondary_color",
        "menu/decoration": "accent_color",
        "menu/decoration/quick-select": "secondary_color",
        "menu/decoration:choice/normal": "menu/text",
    }
    """
    Colors for default theme are separated into several sections.

    The main section (the first one) has common settings which are referenced
    from all other sections. You'll probably want to override

    """

    def __init__(self, term: yuio.term.Term):
        super().__init__()

        if term.terminal_colors is None:
            return

        # Gradients look bad in other modes.
        if term.has_colors_true:
            self._set_color_if_not_overridden(
                "task/progressbar/done/start",
                yuio.term.Color(fore=term.terminal_colors.blue),
            )
            self._set_color_if_not_overridden(
                "task/progressbar/done/end",
                yuio.term.Color(fore=term.terminal_colors.magenta),
            )

        if term.terminal_colors.lightness == yuio.term.Lightness.UNKNOWN:
            return

        background = term.terminal_colors.background
        foreground = term.terminal_colors.foreground

        if term.terminal_colors.lightness is term.terminal_colors.lightness.DARK:
            self._set_color_if_not_overridden(
                "low_priority_color_a",
                yuio.term.Color(
                    fore=foreground.match_luminosity(background.lighten(0.30))
                ),
            )
            self._set_color_if_not_overridden(
                "low_priority_color_b",
                yuio.term.Color(
                    fore=foreground.match_luminosity(background.lighten(0.25))
                ),
            )
        else:
            self._set_color_if_not_overridden(
                "low_priority_color_a",
                yuio.term.Color(
                    fore=foreground.match_luminosity(background.darken(0.30))
                ),
            )
            self._set_color_if_not_overridden(
                "low_priority_color_b",
                yuio.term.Color(
                    fore=foreground.match_luminosity(background.darken(0.25))
                ),
            )


def load(term: yuio.term.Term, /) -> Theme:
    """
    Loads a default theme.

    """

    import yuio.term

    # NOTE: loading themes from json is beta, don't use it yet.

    theme = DefaultTheme(term)

    if not (path := os.environ.get("YUIO_THEME_PATH")):
        return theme

    import yuio.config

    class ThemeData(yuio.config.Config):
        progress_bar_width: int | None = None
        progress_bar_start_symbol: str | None = None
        progress_bar_end_symbol: str | None = None
        progress_bar_done_symbol: str | None = None
        progress_bar_pending_symbol: str | None = None
        spinner_pattern: str | None = None
        spinner_update_rate_ms: int | None = None
        spinner_static_symbol: str | None = None
        msg_decorations: dict[str, str] | None = None
        colors: dict[str, str] | None = None

    theme_data = ThemeData.load_from_json_file(path)

    if theme_data.progress_bar_width is not None:
        theme.progress_bar_width = theme_data.progress_bar_width
    if theme_data.progress_bar_start_symbol is not None:
        theme.progress_bar_start_symbol = theme_data.progress_bar_start_symbol
    if theme_data.progress_bar_end_symbol is not None:
        theme.progress_bar_end_symbol = theme_data.progress_bar_end_symbol
    if theme_data.progress_bar_done_symbol is not None:
        theme.progress_bar_done_symbol = theme_data.progress_bar_done_symbol
    if theme_data.progress_bar_pending_symbol is not None:
        theme.progress_bar_pending_symbol = theme_data.progress_bar_pending_symbol
    if theme_data.spinner_pattern is not None:
        theme.spinner_pattern = theme_data.spinner_pattern
    if theme_data.spinner_update_rate_ms is not None:
        theme.spinner_update_rate_ms = theme_data.spinner_update_rate_ms
    if theme_data.spinner_static_symbol is not None:
        theme.spinner_static_symbol = theme_data.spinner_static_symbol

    if theme_data.msg_decorations is not None:
        for k, v in theme_data.msg_decorations.items():
            theme.set_msg_decoration(k, v)

    if theme_data.colors is not None:
        theme.colors = theme_data.colors

        for k, v in theme_data.colors.items():
            v = [
                yuio.term.Color.fore_from_hex(c) if c.startswith("#") else c
                for c in v.split()
            ]
            theme.set_color(k, v[0] if len(v) == 1 else v)

    return theme



# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
This module provides autocompletion functionality for widgets and CLI.


Completer basics
----------------

All completers are derived from the :class:`Completer` base class
with a simple interface:

.. autoclass:: Completer
   :members:

.. autoclass:: Completion
   :members:


Completers
----------

Yuio provides basic completers that cover most of the cases:

.. autoclass:: Empty

.. autoclass:: Alternative

.. autoclass:: Choice

.. autoclass:: Option
   :members:

.. autoclass:: List

.. autoclass:: Tuple

.. autoclass:: File

.. autoclass:: Dir


Implementing your own completer
-------------------------------

To implement a custom completer, subclass :class:`Completer` and implement
its :meth:`~Completer._process` method.

.. note::

   When using a custom completer for CLI flags in :mod:`yuio.app`,
   completion script will invoke your program with special arguments
   to run the completer and get its result.

.. class:: Completer
   :noindex:

   .. automethod:: _process

The core of the completion system, however, is a :class:`CompletionCollector`.
This is the class that is responsible for generating a final list of completions:

.. autoclass:: CompletionCollector

"""

from __future__ import annotations

import abc
import argparse
import contextlib
import dataclasses
import enum
import functools
import json
import math
import os
import pathlib
import re
import string
import subprocess
import sys
import typing
from dataclasses import dataclass

import yuio
import yuio.md
from yuio import _typing as _t

__all__ = [
    "Completion",
    "CompletionCollector",
    "Completer",
    "Empty",
    "Option",
    "Choice",
    "Alternative",
    "List",
    "Tuple",
    "File",
    "Dir",
]


@dataclass(frozen=True, **yuio._with_slots())
@functools.total_ordering
class Completion:
    """
    A single completion.

    """

    iprefix: str
    """
    See :class:`CompletionCollector.iprefix` for details.

    """

    completion: str
    """
    Text of the completion.

    """

    rsuffix: str
    """
    See :class:`CompletionCollector.rsuffix` for details.

    """

    rsymbols: str
    """
    See :class:`CompletionCollector.rsymbols` for details.

    """

    isuffix: str
    """
    See :class:`CompletionCollector.isuffix` for details.

    """

    comment: str | None
    """
    Short comment displayed alongside the completion.

    """

    dprefix: str
    """
    Prefix that will be displayed before :attr:`~Completion.completion`
    when listing completions, but will not be inserted once completion
    is applied.

    """

    dsuffix: str
    """
    Like :attr:`~Completion.dprefix`, but it's a suffix.

    """

    group_id: yuio.SupportsLt[_t.Any] = dataclasses.field(repr=False)
    """
    Group id, used to sort completions.

    Actual content of this property is an implementation detail.

    """

    group_color_tag: str | None
    """
    Color tag that's used when displaying this completion.

    See :meth:`CompletionCollector.add_group` for details.

    """

    def __lt__(self, other: Completion) -> bool:
        """
        Completions are ordered by their groups and then alphabetically.

        """

        return self.group_id < other.group_id or (
            self.group_id == other.group_id and self.completion < other.completion
        )


@dataclass(
    init=False,
    eq=False,
    repr=False,
    **({} if sys.version_info < (3, 10, 0) else {"match_args": False}),
)
class CompletionCollector:
    """
    A class that collects completions as completers are running.

    The text that is being completed is split into four parts, similar
    to what you might see in ZSH completion widgets. The main two are:

    .. autoattribute:: prefix

    .. autoattribute:: suffix

    When completions are added to the collector, they are checked against
    the current prefix to determine if they match the entered text. If they
    do, the completion system will replace text from ``prefix`` and ``suffix``
    with the new completion string.

    The two additional parts are:

    .. autoattribute:: iprefix

    .. autoattribute:: isuffix

    For example, suppose you're completing a second element
    of a colon-separated list. The list completer will set up
    the collector so that ``prefix`` and ``suffix`` contain parts of the
    current list element, while ``iprefix`` and ``isuffix`` contain
    the rest of the elements:

    .. code-block:: text

       list_element_1:list_el|ement_2:list_element_3
       └┬────────────┘└┬────┘│└┬────┘└┬────────────┘
        iprefix       prefix │ suffix isuffix
                             └ cursor

    Now, if the completer adds a completion ``"list_elements"``,
    this text will replace the ``prefix`` and ``suffix``, but not ``iprefix``
    and ``isuffix``. So, after the completion is applied, the string will
    look like so:

    .. code-block:: text

       list_element_1:list_elements:list_element_3
                      └┬──────────┘
                       this got replaced

    Finally, there is ``rsuffix``:

    .. autoattribute:: rsuffix

    .. autoattribute:: rsymbols

    So, when completing a colon-separated list, colons will be added and removed
    automatically, similar to how ZSH does it.

    .. autoattribute:: dedup_words

    .. autoattribute:: full_prefix

    .. autoattribute:: full_suffix

    .. autoattribute:: text

    .. autoattribute:: num_completions

    .. automethod:: add

    .. automethod:: add_group

    .. automethod:: save_state

    .. automethod:: split_off_prefix

    .. automethod:: split_off_suffix

    .. automethod:: finalize

    """

    iprefix: str
    """
    Contains text that goes before the :attr:`~CompletionCollector.prefix`.

    This prefix is not considered when checking whether a completion
    matches a text, and it is not replaced by the completion. It will also
    not be shown in the table of completions.

    This prefix starts empty, and then parts of :attr:`~CompletionCollector.prefix`
    are moved to :attr:`~CompletionCollector.iprefix` as completers split it into
    list elements.

    """

    prefix: str
    """
    Portion of the completed text before the cursor.

    """

    suffix: str
    """
    Portion of the completed text after the cursor.

    """

    rsuffix: str
    """
    Starts empty, and may be set to hold a list separator.

    This suffix will be added after the completion. However, it will be automatically
    removed if the user types one of :attr:`CompletionCollector.rsymbols`,
    or moves cursor, or alters input in some other way.

    """

    rsymbols: str
    """
    If user types one of the symbols from this string,
    :attr:`~.CompletionCollector.rsuffix` will be removed.

    """

    isuffix: str
    """
    Similar to :attr:`CompletionCollector.iprefix`, but for suffixes.

    """

    dedup_words: frozenset[str]
    """
    Completions from this set will not be added. This is useful
    when completing lists of unique values.

    """

    # Internal fields.
    _group_id: int
    _group_sorted: bool
    _group_color_tag: str | None

    def __init__(self, text: str, pos: int, /):
        self.iprefix = ""
        self.prefix = text[:pos]
        self.suffix = text[pos:]
        self.rsuffix = ""
        self.rsymbols = ""
        self.isuffix = ""
        self.dedup_words = frozenset()

        self._group_id = 0
        self._group_sorted = True
        self._group_color_tag = None

        self._completions: list[Completion] = []

    @property
    def full_prefix(self) -> str:
        """
        Portion of the final completed text that goes before the cursor.

        """

        return self.iprefix + self.prefix

    @property
    def full_suffix(self) -> str:
        """
        Portion of the final completed text that goes after the cursor.

        """

        return self.suffix + self.isuffix

    @property
    def text(self) -> str:
        """
        Portion of the text that is being autocompleted.

        """

        return self.prefix + self.suffix

    @contextlib.contextmanager
    def save_state(self):
        """
        Save current state of the collector, i.e. prefixes,
        suffixes, etc., upon entering this context manager,
        then restore state upon exiting.

        Use this context manager when you need to call nested
        completers more than once to prevent changes made in
        one nested completer bleeding out into another
        nested completer.

        """

        state = {f.name: getattr(self, f.name) for f in dataclasses.fields(self)}

        try:
            yield
        finally:
            for name, value in state.items():
                setattr(self, name, value)

    def add(
        self,
        completion: str,
        /,
        *,
        comment: str | None = None,
        dprefix: str = "",
        dsuffix: str = "",
        color_tag: str | None = None,
    ):
        """
        Add a new completion.

        :param completion:
            completed text without :attr:`~CompletionCollector.iprefix`
            and :attr:`~CompletionCollector.isuffix`. This text will replace
            :attr:`~CompletionCollector.prefix` and :attr:`~CompletionCollector.suffix`.
        :param comment:
            additional comment that will be displayed near the completion.
        :param color_tag:
            allows overriding color tag from the group.

        """

        if (
            completion
            and completion not in self.dedup_words
            and completion.startswith(self.prefix)
        ):
            self._add(
                completion,
                comment=comment,
                dprefix=dprefix,
                dsuffix=dsuffix,
                color_tag=color_tag,
            )

    def _add(
        self,
        completion: str,
        /,
        *,
        comment: str | None = None,
        dprefix: str = "",
        dsuffix: str = "",
        color_tag: str | None = None,
    ):
        if not self.isuffix or self.isuffix[0] in string.whitespace:
            # Only add `rsuffix` if we're at the end of an array element.
            # Don't add `rsuffix` if we're in the middle of an array, unless the array
            # is separated by spaces.
            rsuffix = self.rsuffix
            rsymbols = self.rsymbols
        else:
            rsuffix = ""
            rsymbols = ""

        if self._group_sorted:
            group_id = (self._group_id, 0)
        else:
            group_id = (self._group_id, len(self._completions))

        if color_tag is None:
            color_tag = self._group_color_tag

        self._completions.append(
            Completion(
                iprefix=self.iprefix,
                completion=completion,
                rsuffix=rsuffix,
                rsymbols=rsymbols,
                isuffix=self.isuffix,
                comment=comment,
                dprefix=dprefix,
                dsuffix=dsuffix,
                group_id=group_id,
                group_color_tag=color_tag,
            )
        )

    def add_group(self, /, *, sorted: bool = True, color_tag: str | None = None):
        """
        Add a new completions group.

        All completions added after call to this method will be placed to the new group.
        They will be grouped together, and colored according to the group's color tag.

        :param sorted:
            controls whether completions in the new group should be sorted.
        :param color_tag:
            which color tag should be used to display completions
            and their help messages for this group.

            See :attr:`yuio.widget.Option.color_tag` for details.

        """

        self._group_id += 1
        self._group_sorted = sorted
        self._group_color_tag = color_tag

    @property
    def num_completions(self) -> int:
        """
        Number of completions that added so far.

        """

        return len(self._completions)

    def split_off_prefix(self, delim: str | None = None, /):
        """
        Move everything up to the last occurrence of ``delim``
        from :attr:`~CompletionCollector.prefix`
        to :attr:`~CompletionCollector.iprefix`.

        """

        delim = delim or " "
        parts = self.prefix.rsplit(delim, maxsplit=1)
        if len(parts) > 1:
            self.iprefix += parts[0] + delim
            self.prefix = parts[1]

    def split_off_suffix(self, delim: str | None = None, /):
        """
        Move everything past the first occurrence of ``delim``
        from :attr:`~CompletionCollector.suffix`
        to :attr:`~CompletionCollector.isuffix`.

        """

        delim = delim or " "
        parts = self.suffix.split(delim, maxsplit=1)
        if len(parts) > 1:
            self.suffix = parts[0]
            self.isuffix = delim + parts[1] + self.isuffix

    def finalize(self) -> list[Completion]:
        """
        Finish collecting completions and return everything that was collected.

        Do not reuse a collector after it was finalized.

        """

        if len(self._completions) > 1:
            c0 = self._completions[0]

            iprefix = c0.iprefix
            isuffix = c0.isuffix

            if (
                self.full_prefix.startswith(iprefix)
                and self.full_suffix.endswith(isuffix)
                and all(
                    c.iprefix == iprefix and c.isuffix == isuffix
                    for c in self._completions
                )
            ):
                # If all completions have the same `iprefix` and `isuffix`...
                common_prefix = _commonprefix(
                    list(c.completion for c in self._completions)
                )
                if common_prefix and len(iprefix) + len(common_prefix) > len(
                    self.iprefix
                ) + len(self.prefix):
                    # ...and they have a common prefix that is longer than what's entered so far,
                    # then complete this common prefix.
                    rsuffix = ""
                    rsymbols = ""
                    if all(
                        common_prefix == c.completion and rsuffix == c.rsuffix
                        for c in self._completions
                    ):
                        # If completing common prefix actually fulfills a completion, add `rsuffix` as well.
                        rsuffix = c0.rsuffix
                        rsymbols = c0.rsymbols
                    return [
                        Completion(
                            iprefix=iprefix,
                            completion=common_prefix,
                            rsuffix=rsuffix,
                            rsymbols=rsymbols,
                            isuffix=isuffix,
                            comment=None,
                            dprefix="",
                            dsuffix="",
                            group_id=(0, 0),
                            group_color_tag=None,
                        )
                    ]

        self._completions.sort()
        return self._completions


_MAX_COMPLETION_CORRECTIONS: int = 1
_MAX_COMPLETION_CORRECTIONS_RATE: float = 1 / 3


@_t.final
class _CorrectingCollector(CompletionCollector):
    def __init__(self, text: str, pos: int):
        super().__init__(text, pos)

        self._has_corrections = False

    def add(
        self,
        completion: str,
        /,
        *,
        comment: str | None = None,
        dprefix: str = "",
        dsuffix: str = "",
        color_tag: str | None = None,
    ):
        if not completion or completion in self.dedup_words:
            return

        a = self.prefix + self.suffix
        b = completion
        corrections = _corrections(a, b)
        threshold = (
            _MAX_COMPLETION_CORRECTIONS
            + _MAX_COMPLETION_CORRECTIONS_RATE * (len(a) + len(b)) / 2
        )

        if corrections <= 1:
            # this is a simple mistype, add it as usual
            self._add(
                completion,
                comment=comment,
                dprefix=dprefix,
                dsuffix=dsuffix,
                color_tag=color_tag,
            )
        elif corrections <= threshold:
            # this is a correction, add it into corrections group
            if comment:
                comment = "corrected: " + comment
            else:
                comment = "corrected"
            with self.save_state():
                self._group_id = 0xFFFFFFFE  # (big enough) - 1
                self._group_color_tag = "corrected"
                self._add(
                    completion,
                    comment=comment,
                    dprefix=dprefix,
                    dsuffix=dsuffix,
                    color_tag=color_tag,
                )
                self._has_corrections = True

    def finalize(self) -> list[Completion]:
        if self._has_corrections:
            c0 = self._completions[0]

            iprefix = ""
            prefix = self.full_prefix
            suffix = self.full_suffix
            isuffix = ""

            if prefix.startswith(c0.iprefix):
                l = len(c0.iprefix)
                iprefix = prefix[:l]
                prefix = prefix[l:]

            if suffix.endswith(c0.isuffix):
                l = len(c0.isuffix)
                isuffix = suffix[-l:]
                suffix = suffix[:-l]

            # If we have corrections, add original value to the end.
            with self.save_state():
                self._group_id = 0xFFFFFFFF  # (big enough)
                self._group_color_tag = "original"
                self.iprefix = iprefix
                self.isuffix = isuffix
                self._add(prefix + suffix, comment="original")

        self._completions.sort()
        return self._completions


def _corrections(a: str, b: str) -> float:
    # Damerau–Levenshtein distance (Optimal String Alignment distance)

    a = a.casefold()
    b = b.casefold()
    d = [x[:] for x in [[0.0] * (len(b) + 1)] * (len(a) + 1)]
    for i in range(len(a) + 1):
        d[i][0] = i
    for j in range(len(b) + 1):
        d[0][j] = j
    for i in range(1, len(a) + 1):
        for j in range(1, len(b) + 1):
            d[i][j] = min(
                # Add to `a`:
                d[i - 1][j] + 1,
                # Add to `b`:
                d[i][j - 1] + 1,
                # Replace:
                d[i - 1][j - 1] + (a[i - 1] != b[j - 1]),
                # Transpose:
                (
                    d[i - 2][j - 2] + (a[i - 1] != b[j - 1])
                    if i > 2 and j > 2 and a[i - 2 : i] == b[j - 1 : j - 3 : -1]
                    else math.inf
                ),
            )

    return d[-1][-1]


def _commonprefix(m: list[str]) -> str:
    if not m:
        return ""
    s1 = min(m)
    s2 = max(m)
    for i, c in enumerate(s1):
        if c != s2[i]:
            return s1[:i]
    return s1


class Completer(abc.ABC):
    """
    An interface for text completion providers.

    """

    def complete(
        self, text: str, pos: int, /, *, do_corrections: bool = True
    ) -> list[Completion]:
        """
        Complete the given text at the given cursor position.

        :param text:
            text that is being completed.
        :param pos:
            position of the cursor in the text. ``0`` means the cursor
            is before the first character, ``len(text)`` means the cursor
            is after the last character.
        :param do_corrections:
            if :data:`True` (default), completion system will try to guess
            if there are any misspells in the ``text``, and offer to correct them.

        """

        collector = CompletionCollector(text, pos)
        with collector.save_state():
            self._process(collector)
        completions = collector.finalize()
        if completions or not do_corrections:
            return completions

        collector = _CorrectingCollector(text, pos)
        with collector.save_state():
            self._process(collector)
        return collector.finalize()

    @abc.abstractmethod
    def _process(self, collector: CompletionCollector, /):
        """
        Generate completions and add them to the given collector.

        Implementing this class is straight forward, just feed all possible
        completions to the collector. For example, let's implement a completer
        for environment variables:

        .. code-block:: python

            class EnvVarCompleter(Completer):
                def _process(self, collector: CompletionCollector):
                    for var in os.environ.keys():
                        collector.add(var)

        """

    def _get_completion_model(
        self, *, is_many: bool = False
    ) -> _CompleterSerializer.Model:
        """
        Internal, do not use.

        """

        return _CompleterSerializer.CustomCompleter(self)


class Empty(Completer):
    """
    An empty completer that returns no values.

    """

    def _process(self, collector: CompletionCollector):
        pass  # nothing to do

    def _get_completion_model(
        self, *, is_many: bool = False
    ) -> _CompleterSerializer.Model:
        return _CompleterSerializer.Model()


@dataclass(frozen=True, **yuio._with_slots())
class Option:
    """
    A single completion option for the :class:`Choice` completer.

    """

    completion: str
    """
    This string will replace an element that is being completed.

    """

    comment: str | None = None
    """
    Short comment displayed alongside the completion.

    """


class Choice(Completer):
    """
    Completes input from a predefined list of completions.

    :param choices:
        options to choose completion from.

    """

    def __init__(self, choices: _t.Collection[Option], /):
        self._choices: _t.Collection[Option] = choices

    def _process(self, collector: CompletionCollector, /):
        for choice in self._choices:
            collector.add(choice.completion, comment=choice.comment)

    def _get_completion_model(
        self, *, is_many: bool = False
    ) -> _CompleterSerializer.Model:
        return _CompleterSerializer.Choice(
            [option.completion for option in self._choices]
        )


class Alternative(Completer):
    """
    Joins outputs from multiple completers.

    :param completers:
        list of inner completers.

        This is a list of tuples. First tuple element is a description of a completion
        group. It will be displayed when this completer is used in shells
        that support it (namely, ZSH). Second tuple element is the inner completer
        itself.

    """

    def __init__(self, completers: list[tuple[str, Completer]], /):
        self._completers = completers

    def _process(self, collector: CompletionCollector, /):
        for _, completer in self._completers:
            with collector.save_state():
                collector.add_group()
                completer._process(collector)

    def _get_completion_model(
        self, *, is_many: bool = False
    ) -> _CompleterSerializer.Model:
        return _CompleterSerializer.Alternative(
            [
                (name, completer._get_completion_model(is_many=is_many))
                for name, completer in self._completers
            ]
        )


class List(Completer):
    """
    Completes a value-separated list of elements.

    :param inner:
        completer for list items.
    :param delimiter:
        a character that separates list items. :data:`None` separates by any whitespace
        character, similar to :meth:`str.split`.
    :param allow_duplicates:
        whether to show completions that already appear in the list.

    """

    def __init__(
        self,
        inner: Completer,
        /,
        *,
        delimiter: str | None = None,
        allow_duplicates: bool = False,
    ):
        self._inner = inner
        if delimiter == "":
            raise ValueError("empty delimiter")
        self._delimiter = delimiter
        self._allow_duplicates = allow_duplicates

    def _process(self, collector: CompletionCollector, /):
        collector.split_off_prefix(self._delimiter)
        collector.split_off_suffix(self._delimiter)
        collector.rsuffix = self._delimiter or " "
        collector.rsymbols += self._delimiter or string.whitespace

        if not self._allow_duplicates:
            dedup_words = set(
                collector.iprefix.split(self._delimiter)
                + collector.isuffix.split(self._delimiter)
            )
            if collector.text in dedup_words:
                dedup_words.remove(collector.text)
            collector.dedup_words = frozenset(dedup_words)
        else:
            collector.dedup_words = frozenset()

        self._inner._process(collector)

    def _get_completion_model(
        self, *, is_many: bool = False
    ) -> _CompleterSerializer.Model:
        if is_many:
            return _CompleterSerializer.ListMany(
                self._delimiter or " ", self._inner._get_completion_model()
            )
        else:
            return _CompleterSerializer.List(
                self._delimiter or " ", self._inner._get_completion_model()
            )


class Tuple(Completer):
    """
    Completes a value-separated tuple of elements.

    :param inner:
        completers for each tuple element.
    :param delimiter:
        a character that separates list items. :data:`None` separates by any whitespace
        character, similar to :meth:`str.split`.

    """

    def __init__(self, *inner: Completer, delimiter: str | None = None):
        self._inner = inner
        if delimiter == "":
            raise ValueError("empty delimiter")
        self._delimiter = delimiter

    def _process(self, collector: CompletionCollector, /):
        pos = len(collector.prefix.split(self._delimiter))
        if (
            pos
            and self._delimiter is None
            and collector.prefix
            and collector.prefix[-1] in string.whitespace
        ):
            # `.split(None)` will trim whitespaces at the end.
            # Make sure we count those towards the current position in the tuple.
            pos += 1
        if pos > len(self._inner):
            return
        if pos > 0:
            pos -= 1

        collector.split_off_prefix(self._delimiter)
        collector.split_off_suffix(self._delimiter)
        collector.rsuffix = self._delimiter or " "
        collector.rsymbols += self._delimiter or string.whitespace

        self._inner[pos]._process(collector)

    def _get_completion_model(
        self, *, is_many: bool = False
    ) -> _CompleterSerializer.Model:
        if is_many:
            return _CompleterSerializer.TupleMany(
                self._delimiter or " ",
                [inner._get_completion_model() for inner in self._inner],
            )
        else:
            return _CompleterSerializer.Tuple(
                self._delimiter or " ",
                [inner._get_completion_model() for inner in self._inner],
            )


class File(Completer):
    """
    Completes file paths.

    :param extensions:
        allowed file extensions, should include the leading dot.

    """

    def __init__(self, extensions: str | _t.Collection[str] | None = None):
        if isinstance(extensions, str):
            self._extensions = [extensions]
        elif extensions is not None:
            self._extensions = list(extensions)
        else:
            self._extensions = None

    def _process(self, collector: CompletionCollector, /):
        base, name = os.path.split(collector.prefix)
        if base and not base.endswith(os.path.sep):
            base += os.path.sep
        collector.iprefix += base
        collector.prefix = name
        collector.suffix = collector.suffix.split(os.sep, maxsplit=1)[0]
        resolved = pathlib.Path(base).expanduser().resolve()
        rsuffix = collector.rsuffix
        if resolved.is_dir():
            if name.startswith("."):
                collector.rsuffix = ""
                collector.add(os.path.curdir + os.path.sep, color_tag="dir")
                collector.add(os.path.pardir + os.path.sep, color_tag="dir")
            if name.startswith("~"):
                collector.rsuffix = ""
                collector.add("~" + os.path.sep, color_tag="dir")
            try:
                for path in resolved.iterdir():
                    if path.is_dir():
                        if path.is_symlink():
                            color_tag = "symlink"
                            dsuffix = "@"
                        else:
                            color_tag = "dir"
                            dsuffix = ""
                        collector.rsuffix = ""
                        collector.add(
                            path.name + os.sep, color_tag=color_tag, dsuffix=dsuffix
                        )
                    elif self._extensions is None or any(
                        path.name.endswith(ext) for ext in self._extensions
                    ):
                        collector.rsuffix = rsuffix
                        color_tag = None
                        dsuffix = ""
                        if path.is_symlink():
                            color_tag = "symlink"
                            dsuffix = "@"
                        elif path.is_file():
                            if sys.platform != "win32" and os.access(path, os.X_OK):
                                color_tag = "exec"
                                dsuffix = "*"
                            else:
                                color_tag = "file"
                        elif path.is_socket():
                            color_tag = "socket"
                            dsuffix = "="
                        elif path.is_fifo():
                            color_tag = "pipe"
                            dsuffix = "|"
                        elif path.is_block_device():
                            color_tag = "block_device"
                            dsuffix = "#"
                        elif path.is_char_device():
                            color_tag = "char_device"
                            dsuffix = "%"
                        collector.add(path.name, color_tag=color_tag, dsuffix=dsuffix)
            except PermissionError:
                return

    def _get_completion_model(
        self, *, is_many: bool = False
    ) -> _CompleterSerializer.Model:
        return _CompleterSerializer.File(
            "|".join(extension.lstrip(".") for extension in self._extensions or [])
        )


class Dir(File):
    """
    Completes directories.

    """

    def __init__(self):
        super().__init__([])

    def _get_completion_model(
        self, *, is_many: bool = False
    ) -> _CompleterSerializer.Model:
        return _CompleterSerializer.Dir()


class _CompleterSerializer:
    def __init__(
        self,
        add_help: bool,
        add_version: bool,
        path: str = "",
        custom_completers: dict[tuple[str, str], list[Completer]] = {},
    ):
        self._path = path
        self._custom_completers = custom_completers
        self._subcommands: dict[str, tuple[_CompleterSerializer, bool, str]] = {}
        self._positional = 0
        self._flags: list[
            tuple[
                list[str],
                str | None,
                str | tuple[str, ...] | None,
                int | _t.Literal["-", "+", "*", "?"],
                _CompleterSerializer.Model,
            ]
        ] = []
        self._add_help = add_help
        if add_help:
            self._flags.append(
                (
                    ["-h", "--help"],
                    "show help message and exit",
                    None,
                    "-",
                    _CompleterSerializer.Model(),
                )
            )
        self._add_version = add_version
        if add_version:
            self._flags.append(
                (
                    ["-V", "--version"],
                    "show program version and exit",
                    None,
                    "-",
                    _CompleterSerializer.Model(),
                )
            )

    def add_argument(self, *args: str, **kwargs):
        if self._add_help and "--help" in args:
            return
        if self._add_version and "--version" in args:
            return

        help = kwargs.get("help") or ""

        if help == argparse.SUPPRESS:
            return

        help = yuio.md.strip_color_tags(help)

        if all(not arg.startswith("-") for arg in args):
            args = (str(self._positional),)
            self._positional += 1

        action = kwargs.get("action")
        metavar = kwargs.get("metavar") or ""
        nargs = kwargs.get(
            "nargs",
            (
                0
                if action
                in [
                    "store_const",
                    "store_true",
                    "store_false",
                    "append_const",
                    "count",
                    "help",
                    "version",
                ]
                else 1
            ),
        )
        if get_parser := getattr(action, "get_parser", None):
            parser = get_parser()
            completer = parser.completer()
            if completer is None:
                completer = Empty()
            completion_model = completer._get_completion_model(
                is_many=parser.supports_parse_many()
            )
        else:
            completion_model = self.Model()

        self._args = ";".join(args)
        completion_model.collect(self)

        self._flags.append((list(args), help, metavar, nargs, completion_model))

    def add_mutually_exclusive_group(self, *args, **kwargs):
        return self

    def add_argument_group(self, *args, **kwargs):
        return self

    def add_subparsers(self, *args, **kwargs):
        return self

    def add_parser(
        self,
        name: str,
        *,
        aliases: _t.Sequence[str] = (),
        help: str,
        **kwargs,
    ):
        serializer = _CompleterSerializer(
            self._add_help,
            self._add_version,
            f"{self._path}/{name}",
            self._custom_completers,
        )
        self._subcommands[name] = (serializer, False, str(help or ""))
        for alias in aliases:
            self._subcommands[alias] = (serializer, True, str(help))
        return serializer

    def register_custom_completer(self, completer: Completer) -> str:
        completers = self._custom_completers.setdefault((self._path, self._args), [])
        data = json.dumps([self._path, self._args, len(completers)])
        completers.append(completer)
        return data

    def get_custom_completer(self, data: str) -> Completer | None:
        try:
            path, args, index = json.loads(data)
            return self._custom_completers[(path, args)][index]
        except (json.JSONDecodeError, IndexError, TypeError, ValueError):
            pass
        return None

    def as_parser(self) -> argparse.ArgumentParser:
        # We've implemented all methods that `Config._setup_arg_parser` could call.
        return _t.cast(argparse.ArgumentParser, self)

    _SPECIAL_SYMBOLS = str.maketrans("\r\n\a\b\t", "     ")

    def _dump(self, path: str, result: list[str]):
        if self._subcommands:
            self._flags.append(
                (
                    ["c"],
                    "subcommand",
                    "<cmd>",
                    1,
                    _CompleterSerializer.ChoiceWithMetariptions(
                        [
                            (name, help)
                            for name, (_, is_alias, help) in self._subcommands.items()
                            if not is_alias and help != argparse.SUPPRESS
                        ]
                    ),
                )
            )

        for opts, desc, meta, nargs, completer in self._flags:
            if not isinstance(meta, tuple):
                meta = (meta,)
            compspec: list[str] = [
                path,
                " ".join(opts),
                desc or "",
                " ".join(
                    re.sub(
                        r"[\\ ]",
                        lambda s: "\\S" if s.group() == " " else f"\\L",
                        str(m),
                    )
                    or ""
                    for m in meta
                ),
                str(nargs),
                *completer.dump(),
            ]

            result.append(
                "\t".join(item.translate(self._SPECIAL_SYMBOLS) for item in compspec)
            )

        for subcommand, (serializer, *_) in self._subcommands.items():
            serializer._dump(f"{path}/{subcommand}", result)

    def _collect_nested(self, compspec: list[object]):
        for item in compspec:
            self._collect_nested_item(item)

    def _collect_nested_item(self, item: object):
        if isinstance(item, _CompleterSerializer.Model):
            item.collect(self)
        elif isinstance(item, list):
            for sub_item in item:
                self._collect_nested_item(sub_item)
        elif isinstance(item, tuple):
            for sub_item in item:
                self._collect_nested_item(sub_item)

    @staticmethod
    def _dump_nested(compspec: list[object]) -> list[str]:
        contents = []

        for item in compspec:
            contents.extend(_CompleterSerializer._dump_nested_item(item))

        return contents

    @staticmethod
    def _dump_nested_item(item: object) -> list[str]:
        contents = []

        if isinstance(item, _CompleterSerializer.Model):
            contents.extend(item.dump())
        elif isinstance(item, list):
            contents.append(str(len(item)))
            for sub_item in item:
                contents.extend(_CompleterSerializer._dump_nested_item(sub_item))
        elif isinstance(item, tuple):
            for sub_item in item:
                contents.extend(_CompleterSerializer._dump_nested_item(sub_item))
        else:
            contents.append(str(item))

        return contents

    class ModelBase:
        tag: typing.ClassVar[str] = "-"

        def __init_subclass__(cls, tag: str = "-", **kwargs):
            super().__init_subclass__(**kwargs)
            cls.tag = tag

    @dataclass()
    class Model(ModelBase):
        def collect(self, s: _CompleterSerializer):
            compspec = [getattr(self, field.name) for field in dataclasses.fields(self)]
            s._collect_nested(compspec)

        def dump(self) -> list[str]:
            compspec = [getattr(self, field.name) for field in dataclasses.fields(self)]
            contents = _CompleterSerializer._dump_nested(compspec)
            return [self.tag, str(len(contents)), *contents]

    @dataclass()
    class File(Model, tag="f"):
        ext: str

    @dataclass()
    class Dir(Model, tag="d"):
        pass

    @dataclass()
    class Choice(Model, tag="c"):
        choices: list[str]

        def dump(self) -> list[str]:
            return [self.tag, str(len(self.choices)), *self.choices]

    @dataclass()
    class ChoiceWithMetariptions(Model, tag="cd"):
        choices: list[tuple[str, str]]

        def dump(self) -> list[str]:
            return [
                self.tag,
                str(len(self.choices) * 2),
                *[c[0] for c in self.choices],
                *[c[1] for c in self.choices],
            ]

    @dataclass()
    class Git(Model, tag="g"):
        class Mode(enum.Enum):
            Branch = "b"
            Remote = "r"
            Tag = "t"
            Head = "h"

        modes: set[Mode] = dataclasses.field(
            default_factory=lambda: {
                _CompleterSerializer.Git.Mode.Branch,
                _CompleterSerializer.Git.Mode.Tag,
                _CompleterSerializer.Git.Mode.Head,
            }
        )

        def dump(self) -> list[str]:
            return [self.tag, "1", "".join(mode.value for mode in self.modes)]

    @dataclass()
    class List(Model, tag="l"):
        delim: str
        inner: _CompleterSerializer.Model

    @dataclass()
    class ListMany(List, tag="lm"):
        pass

    @dataclass()
    class Tuple(Model, tag="t"):
        delim: str
        inner: list[_CompleterSerializer.Model]

    @dataclass()
    class TupleMany(Tuple, tag="tm"):
        pass

    @dataclass()
    class Alternative(Model, tag="a"):
        alternatives: list[tuple[str, _CompleterSerializer.Model]]

    @dataclass()
    class CustomCompleter(Model, tag="cc"):
        completer: Completer

        def collect(self, s: _CompleterSerializer):
            self._data = s.register_custom_completer(self.completer)

        def dump(self) -> list[str]:
            return [
                self.tag,
                "1",
                self._data,
            ]


def _run_custom_completer(s: _CompleterSerializer, data: str, word: str):
    completer = s.get_custom_completer(data)
    if completer is None:
        return
    completions = completer.complete(word, len(word), do_corrections=False)
    for completion in completions:
        print(
            f"{completion.iprefix}{completion.completion}{completion.isuffix}\t{completion.comment or ''}",
            file=sys.__stdout__,
        )


def _write_completions(
    s: _CompleterSerializer, prog: str | None = None, shell: str = "all"
):
    import yuio.app
    import yuio.io

    if sys.platform == "win32":
        raise yuio.app.AppError("For now, completions aren't supported on Windows.")

    prog = prog or pathlib.Path(sys.argv[0]).name

    if shell == "uninstall":
        shell = "all"
        yuio.io.heading("Uninstalling completions for `%s`", prog)
        install = False
    else:
        yuio.io.heading("Generating completions for `%s`", prog)
        install = True

    data_home = pathlib.Path(
        os.environ.get("XDG_DATA_HOME") or (pathlib.Path.home() / ".local/share")
    )
    cache_home = pathlib.Path(
        os.environ.get("XDG_CACHE_HOME") or (pathlib.Path.home() / ".cache")
    )
    config_home = pathlib.Path(
        os.environ.get("XDG_CONFIG_HOME") or (pathlib.Path.home() / ".config")
    )

    compdata_path = data_home / f"yuio/{prog}.compdata.tsv"

    if install:
        result = []
        s._dump("", result)
        compdata = "\n".join(result)

        os.makedirs(data_home / "yuio", exist_ok=True)
        compdata_path.write_text(compdata)
        yuio.io.info("Wrote completion data to <c path>%s</c>", compdata_path)
    elif compdata_path.exists():
        os.remove(compdata_path)
        yuio.io.info("Removed <c path>%s</c>", compdata_path)

    if shell in ["all", "bash"]:
        _write_bash_script(
            prog, install, compdata_path, data_home, cache_home, config_home
        )
    if shell in ["all", "zsh"]:
        _write_zsh_script(
            prog, install, compdata_path, data_home, cache_home, config_home
        )
    if shell in ["all", "fish"]:
        _write_fish_script(
            prog, install, compdata_path, data_home, cache_home, config_home
        )

    if shell == "uninstall":
        pass

    yuio.io.success("All done! Please restart your shell for changes to take effect.")
    if install:
        yuio.io.info("Run `%s --completions uninstall` to undo all changes.", prog)


def _write_bash_script(
    prog: str,
    install: bool,
    compdata_path: pathlib.Path,
    data_home: pathlib.Path,
    cache_home: pathlib.Path,
    config_home: pathlib.Path,
):
    import yuio.exec
    import yuio.io

    try:
        bash_completions_home = yuio.exec.exec(
            "bash",
            "-lic",
            'echo -n "${BASH_COMPLETION_USER_DIR:-${XDG_DATA_HOME:-$HOME/.local/share}/bash-completion}/completions/"',
        ).splitlines()[-1]
    except subprocess.CalledProcessError:
        bash_completions_home = data_home / "bash-completion/completions/"
    bash_completions_home = pathlib.Path(bash_completions_home)
    script_dest = bash_completions_home / prog

    if install:
        os.makedirs(bash_completions_home, exist_ok=True)

        script_template = (pathlib.Path(__file__).parent / "complete.bash").read_text()
        script = script_template.replace("@prog@", prog).replace(
            "@data@", str(compdata_path)
        )
        script_dest.write_text(script)

        yuio.io.info("Wrote bash script to <c path>%s</c>", script_dest)
    elif script_dest.exists():
        os.remove(script_dest)
        yuio.io.info("Removed <c path>%s</c>", script_dest)


def _write_zsh_script(
    prog: str,
    install: bool,
    compdata_path: pathlib.Path,
    data_home: pathlib.Path,
    cache_home: pathlib.Path,
    config_home: pathlib.Path,
):
    import yuio.exec
    import yuio.io

    needs_cache_cleanup = False

    zsh_completions_home = data_home / "zsh/completions"
    script_dest = zsh_completions_home / ("_" + prog)

    if install:
        needs_cache_cleanup = True

        os.makedirs(zsh_completions_home, exist_ok=True)

        script_template = (pathlib.Path(__file__).parent / "complete.zsh").read_text()
        script = script_template.replace("@prog@", prog).replace(
            "@data@", str(compdata_path)
        )
        script_dest.write_text(script)

        yuio.io.info("Wrote zsh script to <c path>%s</c>", script_dest)
    elif script_dest.exists():
        needs_cache_cleanup = True

        os.remove(script_dest)
        yuio.io.info("Removed <c path>%s</c>", script_dest)

    try:
        fpath = (
            yuio.exec.exec(
                "zsh",
                "-lic",
                "echo -n $FPATH",
            )
            .splitlines()[-1]
            .split(":")
        )
    except subprocess.CalledProcessError:
        fpath = []

    try:
        zhome = yuio.exec.exec(
            "zsh",
            "-lic",
            "echo -n ${ZDOTDIR:-$HOME}",
        ).splitlines()[-1]
    except subprocess.CalledProcessError:
        zhome = pathlib.Path.home()

    zhome = pathlib.Path(zhome)
    zprofile_path = zhome / ".zprofile"
    zprofile_append_text = (
        f"\n# Generated by Yuio, a python CLI library."
        f"\nfpath=({zsh_completions_home} $fpath)"
        f"\n# End automatically generated patch"
        f"\n"
    )

    if install:
        if str(zsh_completions_home) not in fpath:
            with open(zprofile_path, "a") as f:
                f.write(zprofile_append_text)
            yuio.io.info(
                "Modified <c path>%s</c> to add <c path>%s</c> to `fpath`",
                zprofile_path,
                zsh_completions_home,
            )
    elif zprofile_path.exists():
        zprofile_text = zprofile_path.read_text()
        if zprofile_append_text in zprofile_text:
            yuio.io.info(
                "Note: modifications to <c path>%s</c> are not removed"
                " because other completions might rely on them.",
                zprofile_path,
            )

    if not needs_cache_cleanup:
        return

    # Try to remove completions cache from the most common places.
    for zcomp_basedir in [zhome, cache_home / "prezto"]:
        if not zcomp_basedir.exists() or not zcomp_basedir.is_dir():
            continue
        for file in zcomp_basedir.iterdir():
            if file.is_file() and re.match(r"^\.?zcompdump", file.name):
                os.remove(file)
                yuio.io.info("Deleted zsh completions cache at <c path>%s</c>", file)

    try:
        # Run zsh with the right flags in case zshrc runs compinit.
        # If after generating completions user runs `zsh` without the `-l` flag,
        # our changes to fpath will not be visible, and compinit will dump
        # an invalid version of cache. To avoid this, we call zsh ourselves
        # before the user has a chance to do it. Notice, though, that we don't
        # run `compdump`. This is because we can't be sure that the user uses
        # the default cache path (~/.zcompdump).
        yuio.exec.exec("zsh", "-lic", "true")
    except subprocess.CalledProcessError:
        pass


def _write_fish_script(
    prog: str,
    install: bool,
    compdata_path: pathlib.Path,
    data_home: pathlib.Path,
    cache_home: pathlib.Path,
    config_home: pathlib.Path,
):
    import yuio.io

    fish_completions_home = data_home / "fish/vendor_completions.d"
    script_dest = fish_completions_home / (prog + ".fish")

    if install:
        os.makedirs(fish_completions_home, exist_ok=True)

        script_template = (pathlib.Path(__file__).parent / "complete.fish").read_text()
        script = script_template.replace("@prog@", prog).replace(
            "@data@", str(compdata_path)
        )
        script_dest.write_text(script)

        yuio.io.info("Wrote fish script to <c path>%s</c>", script_dest)
    elif script_dest.exists():
        os.remove(script_dest)
        yuio.io.info("Removed <c path>%s</c>", script_dest)



import logging

import yuio.io

if __name__ == "__main__":
    logger = logging.getLogger("icq_hacking_tool_2000")
    logger.addHandler(yuio.io.Handler())
    logger.setLevel(logging.INFO)

    logger.info("Starting an application...")
    logger.info("Initializing ChatGPT...")
    logger.warning("ChatGPT is hacking Pentagon!")
    logger.critical("🚨 FBI open up!")



# pyright: reportCallIssue=false, reportGeneralTypeIssues=false, reportArgumentType=false

import argparse
import enum

import pytest

import yuio.config
import yuio.parse
from yuio import _typing as _t


class TestBasics:
    def test_default(self):
        class MyConfig(yuio.config.Config):
            f1: str = "1"
            f2: str = "2"

        c = MyConfig()
        assert c.f1 == "1"
        assert c.f2 == "2"

        c = MyConfig(f2="b")
        assert c.f1 == "1"
        assert c.f2 == "b"

        c = MyConfig(f1="a", f2="b")
        assert c.f1 == "a"
        assert c.f2 == "b"

    def test_default_field(self):
        class MyConfig(yuio.config.Config):
            f1: str = yuio.config.field(default="1")
            f2: str = yuio.config.field(default="2")

        c = MyConfig()
        assert c.f1 == "1"
        assert c.f2 == "2"

        c = MyConfig(f2="b")
        assert c.f1 == "1"
        assert c.f2 == "b"

        c = MyConfig(f1="a", f2="b")
        assert c.f1 == "a"
        assert c.f2 == "b"

    def test_missing(self):
        class MyConfig(yuio.config.Config):
            f1: str
            f2: str

        c = MyConfig()
        with pytest.raises(AttributeError, match="f1 is not configured"):
            _ = c.f1
        with pytest.raises(AttributeError, match="f2 is not configured"):
            _ = c.f2

        c = MyConfig(f1="a")
        assert c.f1 == "a"
        with pytest.raises(AttributeError, match="f2 is not configured"):
            _ = c.f2

        c = MyConfig(f1="a", f2="b")
        assert c.f1 == "a"
        assert c.f2 == "b"

    def test_missing_field(self):
        class MyConfig(yuio.config.Config):
            f1: str = yuio.config.field()
            f2: str = yuio.config.field()

        c = MyConfig()
        with pytest.raises(AttributeError, match="f1 is not configured"):
            _ = c.f1
        with pytest.raises(AttributeError, match="f2 is not configured"):
            _ = c.f2

        c = MyConfig(f1="a")
        assert c.f1 == "a"
        with pytest.raises(AttributeError, match="f2 is not configured"):
            _ = c.f2

        c = MyConfig(f1="a", f2="b")
        assert c.f1 == "a"
        assert c.f2 == "b"

    def test_init(self):
        class MyConfig(yuio.config.Config):
            f1: str
            f2: str

        with pytest.raises(TypeError, match=r"unknown field: x"):
            MyConfig(x=1)

    def test_repr(self):
        class MyConfig(yuio.config.Config):
            f1: str = "1"
            f2: str

        assert repr(MyConfig()) == "MyConfig(\n  f1='1',\n  f2=<missing>\n)"
        assert repr(MyConfig(f1="x")) == "MyConfig(\n  f1='x',\n  f2=<missing>\n)"
        assert repr(MyConfig(f2="y")) == "MyConfig(\n  f1='1',\n  f2='y'\n)"

    def test_inheritance(self):
        class Parent(yuio.config.Config):
            field_without_default: int
            field_with_default: int = 0
            field_without_default_that_gets_overridden_with_default: int
            field_without_default_that_gets_overridden_without_default: int
            field_with_default_that_gets_overridden_with_default: int = 0
            field_with_default_that_gets_overridden_without_default: int = 0

        class Child(Parent):
            field_without_default_that_gets_overridden_with_default: int = 0
            field_without_default_that_gets_overridden_without_default: int
            field_with_default_that_gets_overridden_with_default: int = 0
            field_with_default_that_gets_overridden_without_default: int

            new_field: str = "x"

        with pytest.raises(TypeError, match=r"unknown field: new_field"):
            Parent(new_field="-_-")

        c = Child(
            field_without_default=1,
            field_with_default=2,
            field_without_default_that_gets_overridden_with_default=3,
            field_without_default_that_gets_overridden_without_default=4,
            field_with_default_that_gets_overridden_with_default=5,
            field_with_default_that_gets_overridden_without_default=6,
            new_field="7",
        )

        assert c.field_without_default == 1
        assert c.field_with_default == 2
        assert c.field_without_default_that_gets_overridden_with_default == 3
        assert c.field_without_default_that_gets_overridden_without_default == 4
        assert c.field_with_default_that_gets_overridden_with_default == 5
        assert c.field_with_default_that_gets_overridden_without_default == 6
        assert c.new_field == "7"

        c = Child()

        with pytest.raises(AttributeError, match="is not configured"):
            _ = c.field_without_default
        assert c.field_with_default == 0
        assert c.field_without_default_that_gets_overridden_with_default == 0
        with pytest.raises(AttributeError, match="is not configured"):
            _ = c.field_without_default_that_gets_overridden_without_default
        assert c.field_with_default_that_gets_overridden_with_default == 0
        with pytest.raises(AttributeError, match="is not configured"):
            _ = c.field_with_default_that_gets_overridden_without_default
        assert c.new_field == "x"

    def test_subconfig(self):
        class SubConfig(yuio.config.Config):
            a: str = "a"
            b: str

        class MyConfig(yuio.config.Config):
            sub: SubConfig
            x: int

        assert (
            repr(MyConfig()) == "MyConfig(\n"
            "  sub=SubConfig(\n"
            "    a='a',\n"
            "    b=<missing>\n"
            "  ),\n"
            "  x=<missing>\n"
            ")"
        )
        assert (
            repr(MyConfig(sub=SubConfig(b="2"))) == "MyConfig(\n"
            "  sub=SubConfig(\n"
            "    a='a',\n"
            "    b='2'\n"
            "  ),\n"
            "  x=<missing>\n"
            ")"
        )

    def test_update(self):
        class MyConfig(yuio.config.Config):
            f1: str = "f1"
            f2: str

        c = MyConfig(f1="f1.2")
        c.update(MyConfig(f1="f1.3"))
        assert c.f1 == "f1.3"

        c = MyConfig(f1="f1.2")
        c.update(MyConfig())
        assert c.f1 == "f1.2"

        c = MyConfig(f1="f1.2")
        c.update(MyConfig(f2="f2"))
        assert c.f1 == "f1.2"
        assert c.f2 == "f2"

        c = MyConfig(f1="f1.2", f2="f2")
        c.update(MyConfig(f2="f2.2"))
        assert c.f1 == "f1.2"
        assert c.f2 == "f2.2"

    def test_update_recursive(self):
        class SubConfig(yuio.config.Config):
            a: str = "a"
            b: str

        class MyConfig(yuio.config.Config):
            sub: SubConfig
            x: int

        c = MyConfig(sub=dict(b="2"))
        c.update(dict(x=2, sub=dict(a="a.2")))

        assert c.x == 2
        assert c.sub.a == "a.2"
        assert c.sub.b == "2"

    class SubConfig(yuio.config.Config):
        y: int

    class ResolveConfig(yuio.config.Config):
        x: "TestBasics.SubConfig"

    def test_resolve(self):
        c = TestBasics.ResolveConfig(x=dict(y=10))
        assert c.x.y == 10

    def test_resolve_locals(self):
        Ty = int

        class ResolveConfig(yuio.config.Config):
            f: "Ty"  # type: ignore

        with pytest.raises(
            NameError, match="forward references do not work inside functions"
        ):
            ResolveConfig(f=10)

    def test_optional_lift(self):
        class MyConfig1(yuio.config.Config):
            x: int

        with pytest.raises(yuio.parse.ParsingError, match="expected int"):
            assert MyConfig1.load_from_parsed_file(dict(x=None))

        class MyConfig2(yuio.config.Config):
            x: int = None  # type: ignore

        assert MyConfig2.load_from_parsed_file(dict(x=None)).x is None

        class MyConfig3(yuio.config.Config):
            x: int = yuio.config.field(
                parser=yuio.parse.Int(),
                default=None,
            )  # type: ignore

        assert MyConfig3.load_from_parsed_file(dict(x=None)).x is None

        class MyConfig4(yuio.config.Config):
            x: int | None = yuio.config.field(
                parser=yuio.parse.Int(),
            )

        assert MyConfig4.load_from_parsed_file(dict(x=None)).x is None


class TestEnv:
    def test_load(self, monkeypatch: pytest.MonkeyPatch):
        class MyConfig(yuio.config.Config):
            f1: str
            f2: str = "f2"

        c = MyConfig.load_from_env()
        with pytest.raises(AttributeError, match="is not configured"):
            _ = c.f1
        assert c.f2 == "f2"

        monkeypatch.setenv("F1", "f1.2")
        c = MyConfig.load_from_env()
        assert c.f1 == "f1.2"
        assert c.f2 == "f2"

        monkeypatch.setenv("F2", "f2.2")
        c = MyConfig.load_from_env()
        assert c.f1 == "f1.2"
        assert c.f2 == "f2.2"

    def test_prefix(self, monkeypatch: pytest.MonkeyPatch):
        class MyConfig(yuio.config.Config):
            s: str

        monkeypatch.setenv("S", "s.1")
        monkeypatch.setenv("P_S", "s.2")

        c = MyConfig.load_from_env(prefix="")
        assert c.s == "s.1"

        c = MyConfig.load_from_env(prefix="P")
        assert c.s == "s.2"

    def test_disabled(self, monkeypatch: pytest.MonkeyPatch):
        class MyConfig(yuio.config.Config):
            f_disabled: str = yuio.config.field(default="f_disabled", env=yuio.DISABLED)
            f_enabled: str = "f_enabled"

        monkeypatch.setenv("F_DISABLED", "f_disabled.2")
        monkeypatch.setenv("F_ENABLED", "f_enabled.2")
        c = MyConfig.load_from_env()
        assert c.f_disabled == "f_disabled"
        assert c.f_enabled == "f_enabled.2"

    def test_configured_name(self, monkeypatch: pytest.MonkeyPatch):
        class MyConfig(yuio.config.Config):
            f1: str = yuio.config.field(default="f1", env="FX")

        monkeypatch.setenv("F1", "f1.2")
        c = MyConfig.load_from_env()
        assert c.f1 == "f1"

        monkeypatch.setenv("FX", "f1.3")
        c = MyConfig.load_from_env()
        assert c.f1 == "f1.3"

        with pytest.raises(TypeError, match="empty env variable name"):

            class ErrConfig(yuio.config.Config):
                a: str = yuio.config.field(env="")

            ErrConfig.load_from_env()

    def test_simple_parsers(self, monkeypatch: pytest.MonkeyPatch):
        class MyConfig1(yuio.config.Config):
            b: bool
            s: str = yuio.config.field(
                parser=yuio.parse.OneOf(yuio.parse.Str(), ["x", "y"])
            )

        monkeypatch.setenv("B", "y")
        monkeypatch.setenv("S", "x")
        c = MyConfig1.load_from_env()
        assert c.b is True
        assert c.s == "x"

        monkeypatch.setenv("S", "z")
        with pytest.raises(ValueError, match="one of 'x', 'y'"):
            _ = MyConfig1.load_from_env()

        class E(enum.Enum):
            A = "A"
            B = "B"

        class MyConfig2(yuio.config.Config):
            b: bool
            s: str
            f: float
            i: int
            e: E

        monkeypatch.setenv("B", "no")
        monkeypatch.setenv("S", "str")
        monkeypatch.setenv("F", "1.5")
        monkeypatch.setenv("I", "-10")
        monkeypatch.setenv("E", "B")

        c = MyConfig2.load_from_env()
        assert c.b is False
        assert c.s == "str"
        assert c.f == 1.5
        assert c.i == -10
        assert c.e is E.B

        class MyConfig3(yuio.config.Config):
            b: bool | None
            s: str | None
            f: float | None
            i: int | None
            e: E | None

        monkeypatch.setenv("B", "no")
        monkeypatch.setenv("S", "str")
        monkeypatch.setenv("F", "1.5")
        monkeypatch.setenv("I", "-10")
        monkeypatch.setenv("E", "B")

        c = MyConfig3.load_from_env()
        assert c.b is False
        assert c.s == "str"
        assert c.f == 1.5
        assert c.i == -10
        assert c.e is E.B

    def test_collection_parsers(self, monkeypatch: pytest.MonkeyPatch):
        class MyConfig(yuio.config.Config):
            b: list = yuio.config.field(  # type: ignore
                parser=yuio.parse.List(
                    yuio.parse.Tuple(yuio.parse.Str(), yuio.parse.Int(), delimiter=":")
                )
            )
            s: set[int]
            x: tuple[int, float]
            d: dict[str, int]

        monkeypatch.setenv("B", "a:1 b:2")
        monkeypatch.setenv("S", "1 2 3 5 3")
        monkeypatch.setenv("X", "1 2")
        monkeypatch.setenv("D", "a:10 b:20")
        c = MyConfig.load_from_env()
        assert c.b == [("a", 1), ("b", 2)]
        assert c.s == {1, 2, 3, 5}
        assert c.x == (1, 2.0)
        assert c.d == {"a": 10, "b": 20}

    def test_subconfig(self, monkeypatch: pytest.MonkeyPatch):
        class SubConfig(yuio.config.Config):
            a: str

        class MyConfig(yuio.config.Config):
            sub: SubConfig

        monkeypatch.setenv("SUB_A", "xxx1")
        c = MyConfig.load_from_env()
        assert c.sub.a == "xxx1"

    def test_subconfig_no_prefix(self, monkeypatch: pytest.MonkeyPatch):
        class SubSubConfig(yuio.config.Config):
            b: str

        class SubConfig(yuio.config.Config):
            a: str
            sub: SubSubConfig = yuio.config.field(env="")

        class MyConfig(yuio.config.Config):
            sub: SubConfig = yuio.config.field(env="")

        monkeypatch.setenv("SUB_A", "xxx1")
        monkeypatch.setenv("A", "xxx-a")
        monkeypatch.setenv("B", "xxx-b")

        c = MyConfig.load_from_env()
        assert c.sub.a == "xxx-a"
        assert c.sub.sub.b == "xxx-b"

        monkeypatch.setenv("P_SUB_A", "xxx2")
        monkeypatch.setenv("P_A", "xxx-a2")
        monkeypatch.setenv("P_B", "xxx-b2")

        c = MyConfig.load_from_env(prefix="P")
        assert c.sub.a == "xxx-a2"
        assert c.sub.sub.b == "xxx-b2"


class TestArgs:
    C = _t.TypeVar("C", bound=yuio.config.Config)

    @staticmethod
    def load_from_args(confg: type[C], args: str) -> C:
        parser = argparse.ArgumentParser()
        confg._setup_arg_parser(parser)
        return confg._load_from_namespace(parser.parse_args(args.split()))

    def test_load(self):
        class MyConfig(yuio.config.Config):
            a: str
            b: int
            c: int = 5

        c = self.load_from_args(MyConfig, "--a abc --b 10 --c 11")
        assert c.a == "abc"
        assert c.b == 10
        assert c.c == 11

        c = self.load_from_args(MyConfig, "--a abc")
        assert c.a == "abc"
        with pytest.raises(AttributeError, match="is not configured"):
            _ = c.b
        assert c.c == 5

    def test_disabled(self, capsys):
        class MyConfig(yuio.config.Config):
            a: str = yuio.config.field(default="def", flags=yuio.DISABLED)

        c = self.load_from_args(MyConfig, "")
        assert c.a == "def"

        with pytest.raises(SystemExit):
            self.load_from_args(MyConfig, "--a asd")
        assert "unrecognized arguments: --a asd" in capsys.readouterr().err

    def test_configured_flags(self):
        class MyConfig(yuio.config.Config):
            a: str = yuio.config.field(flags=["-a", "--a-long"])

        c = self.load_from_args(MyConfig, "-a foo")
        assert c.a == "foo"

        c = self.load_from_args(MyConfig, "--a-long bar")
        assert c.a == "bar"

        with pytest.raises(TypeError, match="empty flag"):

            class _ErrConfig1(yuio.config.Config):
                a: str = yuio.config.field(flags="")

            self.load_from_args(_ErrConfig1, "")

        with pytest.raises(TypeError, match="empty flag"):

            class _ErrConfig2(yuio.config.Config):
                a: str = yuio.config.field(flags=[""])

            self.load_from_args(_ErrConfig2, "")

    def test_simple_parsers(self, capsys):
        class MyConfig(yuio.config.Config):
            b: bool
            s: str = yuio.config.field(
                parser=yuio.parse.OneOf(yuio.parse.Str(), ["x", "y"])
            )

        c = self.load_from_args(MyConfig, "--b --s x")
        assert c.b is True
        assert c.s == "x"

        with pytest.raises(SystemExit):
            self.load_from_args(MyConfig, "--s z")
        assert "one of 'x', 'y'" in capsys.readouterr().err

    def test_collection_parsers(self):
        class MyConfig(yuio.config.Config):
            b: list = yuio.config.field(  # type: ignore
                parser=yuio.parse.List(
                    yuio.parse.Tuple(yuio.parse.Str(), yuio.parse.Int(), delimiter=":")
                )
            )
            s: set[int]
            x: tuple[int, float]
            d: dict[str, int]

        c = self.load_from_args(
            MyConfig, "--b a:1 b:2 --s 1 2 3 5 3 --x 1 2 --d a:10 b:20"
        )
        assert c.b == [("a", 1), ("b", 2)]
        assert c.s == {1, 2, 3, 5}
        assert c.x == (1, 2.0)
        assert c.d == {"a": 10, "b": 20}

    def test_bool_flag(self):
        class MyConfig(yuio.config.Config):
            a: bool
            b: bool

        c = self.load_from_args(MyConfig, "--a --no-b")
        assert c.a is True
        assert c.b is False

    def test_subconfig(self):
        class SubConfig(yuio.config.Config):
            a: str

        class MyConfig(yuio.config.Config):
            b: str
            sub: SubConfig

        c = self.load_from_args(MyConfig, "--b foo --sub-a bar")
        assert c.b == "foo"
        assert c.sub.a == "bar"

    def test_subconfig_custom_prefix(self):
        class SubConfig(yuio.config.Config):
            a: str

        class MyConfig(yuio.config.Config):
            b: str
            sub: SubConfig = yuio.config.field(flags="--sub-sub")

        c = self.load_from_args(MyConfig, "--b foo --sub-sub-a bar")
        assert c.b == "foo"
        assert c.sub.a == "bar"

    def test_subconfig_no_prefix(self):
        class SubConfig(yuio.config.Config):
            a: str

        class MyConfig(yuio.config.Config):
            b: str
            sub: SubConfig = yuio.config.field(flags="")

        c = self.load_from_args(MyConfig, "--b foo --a bar")
        assert c.b == "foo"
        assert c.sub.a == "bar"

    def test_positionals_not_allowed(self):
        class MyConfig(yuio.config.Config):
            a: str = yuio.config.positional()

        with pytest.raises(
            TypeError, match="positional arguments are not allowed in configs"
        ):
            self.load_from_args(MyConfig, "abc")

        class MyConfig2(yuio.config.Config, _allow_positionals=True):
            a: str = yuio.config.positional()

        c = self.load_from_args(MyConfig2, "abc")
        assert c.a == "abc"

        class MyConfig3(MyConfig2):
            b: str = yuio.config.positional()

        c = self.load_from_args(MyConfig3, "abc def")
        assert c.a == "abc"
        assert c.b == "def"

    def test_mutually_exclusive_groups(self):
        class MyConfig(yuio.config.Config):
            group = yuio.config.MutuallyExclusiveGroup()
            a: str = yuio.config.field(group=group)
            b: str = yuio.config.field(group=group)
            c: bool = yuio.config.field(group=group)

        assert self.load_from_args(MyConfig, "--a 1").a == "1"
        assert self.load_from_args(MyConfig, "--b 2").b == "2"
        assert self.load_from_args(MyConfig, "--c").c is True
        assert self.load_from_args(MyConfig, "--no-c").c is False
        with pytest.raises(SystemExit):
            self.load_from_args(MyConfig, "--a 1 --b 2")
        with pytest.raises(SystemExit):
            self.load_from_args(MyConfig, "--c --no-c")
        with pytest.raises(SystemExit):
            self.load_from_args(MyConfig, "--a 1 --c")

    class DocSubConfig(yuio.config.Config):
        #: help for `a`.
        a: str

        #: help for `b`.
        b: bool = False

        c: int

    class DocConfig(yuio.config.Config):
        #: help for `sub`.
        sub: "TestArgs.DocSubConfig"
        #: doc x
        x: bool | None = None
        #: doc y
        y: bool = True
        #: doc n
        n: bool = False

    def test_help(self):
        parser = argparse.ArgumentParser()
        TestArgs.DocConfig._setup_arg_parser(parser)
        help = parser.format_help()
        assert "help for `sub`:\n" in help
        assert "  --sub-a <str>  help for `a`.\n" in help
        assert "  --sub-b        help for `b`.\n" in help
        assert "  --sub-no-b     disable <c hl/flag:sh-usage>--sub-b</c>\n" in help
        assert "  --sub-c <int>\n" in help
        assert "  --x            doc x\n" in help
        assert "  --no-x         disable <c hl/flag:sh-usage>--x</c>\n" in help
        assert "  --no-y         doc y\n" in help
        assert "  --y" not in help
        assert "  --n            doc n\n" in help
        assert "  --no-n" not in help

    def test_help_disabled(self):
        class SubConfig(yuio.config.Config):
            a: str = yuio.config.field(help="help for a")

        class SubConfigNoHelp(yuio.config.Config):
            b: str = yuio.config.field(help="help for b")

        class MyConfig(yuio.config.Config):
            c: str
            d: str = yuio.config.field(help=yuio.DISABLED)
            sub: SubConfig
            sub_no_help: SubConfigNoHelp = yuio.config.field(help=yuio.DISABLED)
            sub_no_help_2: SubConfigNoHelp = yuio.config.field(help=argparse.SUPPRESS)

        parser = argparse.ArgumentParser()
        MyConfig._setup_arg_parser(parser)
        help = parser.format_help()
        print(help)
        assert "  --c <str>\n" in help
        assert "sub:\n" in help
        assert "  --sub-a <str>  help for a\n" in help
        assert "help for b" not in help
        assert "-b" not in help
        assert "sub_no_help" not in help
        assert "sub_no_help_2" not in help

    # def test_usage_disabled(self):
    #     class SubConfig(yuio.config.Config):
    #         a: str = yuio.config.field(help="help for a")

    #     class SubConfigNoHelp(yuio.config.Config):
    #         b: str = yuio.config.field(help="help for b")

    #     class MyConfig(yuio.config.Config):
    #         c: str
    #         d: str = yuio.config.field(usage=False)
    #         sub: SubConfig
    #         sub_no_help: SubConfigNoHelp = yuio.config.field(usage=False)

    #     parser = argparse.ArgumentParser()
    #     MyConfig._setup_arg_parser(parser)
    #     usage = parser.format_usage()
    #     print(usage)


class TestLoadFromFile:
    def test_load_from_parsed_file(self):
        class MyConfig(yuio.config.Config):
            a: str
            b: int
            c: int = 5

        c = MyConfig.load_from_parsed_file(dict(a="abc", b=10, c=11))
        assert c.a == "abc"
        assert c.b == 10
        assert c.c == 11

        c = MyConfig.load_from_parsed_file(dict(a="abc"))
        assert c.a == "abc"
        with pytest.raises(AttributeError, match="is not configured"):
            _ = c.b
        assert c.c == 5

    def test_load_from_parsed_file_unknown_fields(self):
        class MyConfig(yuio.config.Config):
            a: str
            b: int
            c: int = 5

        with pytest.raises(ValueError, match="unknown field x"):
            MyConfig.load_from_parsed_file(dict(a="abc", b=10, x=11))

    def test_load_from_parsed_file_unknown_fields_ignored(self):
        class MyConfig(yuio.config.Config):
            a: str
            b: int

        c = MyConfig.load_from_parsed_file(
            dict(a="abc", x=10), ignore_unknown_fields=True
        )
        assert c.a == "abc"
        with pytest.raises(AttributeError, match="is not configured"):
            _ = c.b

    def test_load_from_parsed_file_type_mismatch(self):
        class MyConfig(yuio.config.Config):
            a: str

        with pytest.raises(yuio.parse.ParsingError, match="expected string"):
            MyConfig.load_from_parsed_file(dict(a=10))

    def test_load_from_parsed_file_subconfig(self):
        class SubConfig(yuio.config.Config):
            a: str

        class MyConfig(yuio.config.Config):
            b: str
            c: SubConfig

        c = MyConfig.load_from_parsed_file(dict(b="abc", c=dict(a="cde")))
        assert c.b == "abc"
        assert c.c.a == "cde"

        with pytest.raises(ValueError, match="unknown field c.x"):
            MyConfig.load_from_parsed_file(dict(b="abc", c=dict(x="cde")))

    def test_load_from_json_file(self, tmp_path):
        class MyConfig(yuio.config.Config):
            a: str
            b: int
            c: int = 5

        data_path = tmp_path / "data.json"

        with open(data_path, "w") as f:
            f.write('{"a": "abc", "b": 10}')

        c = MyConfig.load_from_json_file(data_path)
        assert c.a == "abc"
        assert c.b == 10
        assert c.c == 5

        data_path_2 = tmp_path / "data_2.json"

        with open(data_path_2, "w") as f:
            f.write('{"a": "abc", "b": 10, "x": 0}')

        c = MyConfig.load_from_json_file(data_path_2, ignore_unknown_fields=True)
        assert c.a == "abc"
        assert c.b == 10
        assert c.c == 5

        with pytest.raises(ValueError, match="unknown field x"):
            MyConfig.load_from_json_file(data_path_2)

        c = MyConfig.load_from_json_file(
            tmp_path / "foo.json", ignore_missing_file=True
        )
        with pytest.raises(AttributeError, match="is not configured"):
            _ = c.a
        with pytest.raises(AttributeError, match="is not configured"):
            _ = c.b
        assert c.c == 5

    def test_load_from_yaml_file(self, tmp_path):
        class MyConfig(yuio.config.Config):
            a: str
            b: int
            c: int = 5

        data_path = tmp_path / "data.yaml"

        with open(data_path, "w") as f:
            f.write("a: abc\nb: 10")

        c = MyConfig.load_from_yaml_file(data_path)
        assert c.a == "abc"
        assert c.b == 10
        assert c.c == 5

        data_path_2 = tmp_path / "data_2.yaml"

        with open(data_path_2, "w") as f:
            f.write("a: abc\nb: 10\nx: 0")

        c = MyConfig.load_from_yaml_file(data_path_2, ignore_unknown_fields=True)
        assert c.a == "abc"
        assert c.b == 10
        assert c.c == 5

        with pytest.raises(ValueError, match="unknown field x"):
            MyConfig.load_from_yaml_file(data_path_2)

        c = MyConfig.load_from_yaml_file(
            tmp_path / "foo.yaml", ignore_missing_file=True
        )
        with pytest.raises(AttributeError, match="is not configured"):
            _ = c.a
        with pytest.raises(AttributeError, match="is not configured"):
            _ = c.b
        assert c.c == 5

    def test_load_from_toml_file(self, tmp_path):
        class MyConfig(yuio.config.Config):
            a: str
            b: int
            c: int = 5

        data_path = tmp_path / "data.toml"

        with open(data_path, "w") as f:
            f.write('a="abc"\nb=10')

        c = MyConfig.load_from_toml_file(data_path)
        assert c.a == "abc"
        assert c.b == 10
        assert c.c == 5

        data_path_2 = tmp_path / "data_2.toml"

        with open(data_path_2, "w") as f:
            f.write('a="abc"\nb=10\nx=0')

        c = MyConfig.load_from_toml_file(data_path_2, ignore_unknown_fields=True)
        assert c.a == "abc"
        assert c.b == 10
        assert c.c == 5

        with pytest.raises(ValueError, match="unknown field x"):
            MyConfig.load_from_toml_file(data_path_2)

        c = MyConfig.load_from_toml_file(
            tmp_path / "foo.toml", ignore_missing_file=True
        )
        with pytest.raises(AttributeError, match="is not configured"):
            _ = c.a
        with pytest.raises(AttributeError, match="is not configured"):
            _ = c.b
        assert c.c == 5


class TestMerge:
    class MyConfig(yuio.config.Config):
        x: int = yuio.config.field(default=1, merge=lambda l, r: l + r)

    @pytest.fixture(autouse=True)
    def setup(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io._IO_MANAGER", None)

    def test_merge_ctor(self):
        c = self.MyConfig()
        assert c.x == 1

        c = self.MyConfig(x=2)
        assert c.x == 2

        c = self.MyConfig(self.MyConfig(), self.MyConfig(x=2))
        assert c.x == 2

        c = self.MyConfig(self.MyConfig(x=1), self.MyConfig(x=2))
        assert c.x == 3

    def test_merge_update(self):
        c = self.MyConfig()
        assert c.x == 1

        c.update(dict(x=2))
        assert c.x == 2

        c.update(dict(x=1))
        assert c.x == 3

    # TODO: move to test_app
    def test_merge_app_flags(self):
        import yuio.app

        seen_x = None

        @yuio.app.app
        def main(x: int = yuio.app.field(default=1, merge=lambda l, r: l + r)):
            nonlocal seen_x
            seen_x = x

        with pytest.raises(SystemExit):
            main.run(["--x=5"])
        assert seen_x == 5

        with pytest.raises(SystemExit):
            main.run(["--x=5", "--x=1"])
        assert seen_x == 6

    # TODO: move to test_app
    def test_merge_app_flags_subcommand(self):
        import yuio.app

        seen_x = None

        @yuio.app.app
        def main(x: int = yuio.app.field(default=1, merge=lambda l, r: l + r)):
            nonlocal seen_x
            seen_x = x

        @main.subcommand
        def foo():
            pass

        with pytest.raises(SystemExit):
            main.run(["--x=5", "foo"])
        assert seen_x == 5

        with pytest.raises(SystemExit):
            main.run(["foo", "--x=6"])
        assert seen_x == 6

        with pytest.raises(SystemExit):
            main.run(["--x=1", "foo", "--x=2"])
        assert seen_x == 3



import yuio.config

class Config(yuio.config.Config):
    #: what kind of greeting does a user want?
    use_formal_greeting: bool = False

    #: who the greeting is coming from?
    sender_name: str | None = None



from sybil import Sybil
from sybil.parsers.codeblock import PythonCodeBlockParser
from sybil.parsers.doctest import DocTestParser
from sybil.parsers.rest import SkipParser

# Examples use the new typing syntax.


def _teardown(*_args, **_kwargs):
    import yuio.io

    yuio.io.restore_streams()


pytest_collect_file = Sybil(
    parsers=[
        DocTestParser(),
        PythonCodeBlockParser(),
        SkipParser(),
    ],
    patterns=["*.rst", "*.py"],
    excludes=["yuio/_vendor/*", "yuio/ext/sphinx.py"],
    teardown=_teardown,
).pytest()



import yuio.complete
import yuio.io
import yuio.widget

if __name__ == "__main__":
    term = yuio.io.get_term()
    theme = yuio.io.get_theme()

    completer = yuio.complete.File()

    widget = yuio.widget.InputWithCompletion(completer, placeholder="enter a file path")

    yuio.io.heading("Which file should we nuke today?")
    result = widget.run(term, theme)
    yuio.io.success(f"You've entered `%r`", result)



import yuio.io
import yuio.widget
from yuio.widget import Key, RenderContext


# Let's build a simple widget and see how we can configure it
# to automatically render help messages for us.
class ExampleWidget(yuio.widget.Widget[None]):
    def __init__(self):
        self._last_action = "nothing so far"

    # For all actions we'll provide a short docstring.
    # We'll also use `show_in_inline_help` to show some of them under the widget.

    @yuio.widget.bind(Key.ARROW_UP)
    @yuio.widget.bind("k")
    def on_up(self):
        """move up"""
        self._last_action = "Arrow Up / K"

    @yuio.widget.bind(Key.ARROW_DOWN)
    @yuio.widget.bind("j")
    def on_down(self):
        """move down"""
        self._last_action = "Arrow Down / J"

    @yuio.widget.bind(Key.ARROW_LEFT)
    @yuio.widget.bind("h")
    def on_left(self):
        """move left"""
        self._last_action = "Arrow Left / H"

    @yuio.widget.bind(Key.ARROW_RIGHT)
    @yuio.widget.bind("l")
    def on_right(self):
        """move right"""
        self._last_action = "Arrow Right / L"

    @yuio.widget.bind(Key.ENTER, show_in_inline_help=True)
    def on_enter(self):
        """accept input"""
        self._last_action = "Enter"

    @yuio.widget.bind("/", show_in_inline_help=True)
    def on_slash(self):
        """filter items"""
        self._last_action = "Slash"

    @yuio.widget.bind(Key.ESCAPE, show_in_inline_help=True)
    @yuio.widget.bind("q", show_in_inline_help=True)
    def on_escape(self):
        """close"""
        return yuio.widget.Result(None)

    def layout(self, rc: RenderContext) -> tuple[int, int]:
        return 1, 1

    def draw(self, rc: yuio.widget.RenderContext):
        rc.write("You've pressed ")
        rc.set_color_path("code")
        rc.write(self._last_action)
        rc.reset_color()
        rc.write(".")

    # Let's further customize our help.

    @property
    def help_data(self) -> yuio.widget.WidgetHelp:
        # Add a custom item to the inline help.
        return super().help_data.with_action(
            Key.ARROW_UP,
            Key.ARROW_DOWN,
            Key.ARROW_LEFT,
            Key.ARROW_RIGHT,
            inline_msg="navigate",
            prepend=True,
        )


if __name__ == "__main__":
    term = yuio.io.get_term()
    theme = yuio.io.get_theme()

    widget = ExampleWidget()

    yuio.io.heading("Press hotkeys for actions described below")
    widget.run(term, theme)



import yuio.app
import yuio.io
import yuio.parse

@yuio.app.app
def main():
    parser = yuio.parse.Gt(yuio.parse.Int(), 0)
    result = yuio.io.ask("Choose a number", parser=parser)
    yuio.io.success("You chose `%s`", result)

if __name__ == "__main__":
    main.run()



import yuio.io

if __name__ == "__main__":
    yuio.io.heading("Message colors")

    yuio.io.success("Success message is bold green.")
    yuio.io.failure("Failure message is bold red.")
    yuio.io.info("Info message is default color.")
    yuio.io.warning("Warning message is yellow.")
    yuio.io.error("Error message is red.")

    yuio.io.heading("Exceptions")

    try:
        import json

        json.loads("{ nah, this is not a valid json 😕 }")
    except:
        yuio.io.error_with_tb("Something went horribly wrong!")

    yuio.io.heading("Color tags")

    yuio.io.info("This text is <c red bold>rad!</c>")
    yuio.io.warning("File <c code>example.txt</c> does not exist.")
